/*
 * SSD1306_I2C.h
 *
 * Created: 2017-04-01
 * version: 2017-04-01
 *  Author: uwezi
 */


// display routines for SSD1306-based OLED displays

#include <stdlib.h>
#include <util/atomic.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "SSD1306_I2C.h"

#ifndef F_CPU
#define F_CPU 20000000UL   // save to assume the fastest clock
#endif

#include <util/delay.h>
#include "i2cmaster.h"

//  6x8 font
//    LSB is top
//    MSB is bottom
//
static const uint8_t smallFont[][6] PROGMEM =
#include "font_6x8_iso8859_1.h"

//
//  define a local copy of the display memory
//
uint8_t framebuffer[DISPLAYSIZEX*DISPLAYSIZEY/8];

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_clearbuffer
  Description  :  Clears the framebuffer but does not transfer to LCD.
  Argument(s)  :  None.
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_clearbuffer(void)
{
  memset(framebuffer, 0x00, DISPLAYSIZEX*DISPLAYSIZEY/8);
}

void SSD1306_sendCommand1(uint8_t command)
{
  i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
  i2c_write(0x00);
  i2c_write(command);
  i2c_stop();
}

void SSD1306_sendCommand2(uint8_t command, uint8_t val1)
{
  i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
  i2c_write(0x00);
  i2c_write(command);
  i2c_write(val1);
  i2c_stop();
}

void SSD1306_sendCommand3(uint8_t command, uint8_t val1, uint8_t val2)
{
  i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
  i2c_write(0x00);
  i2c_write(command);
  i2c_write(val1);
  i2c_write(val2);
  i2c_stop();
}

void SSD1306_sendCommandList(const uint8_t *commands, uint8_t count)
{
  uint8_t bytesOut = 1;
  i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
  i2c_write(0x00);
  
  while (count--) 
  {
    if (bytesOut >= WIRE_MAX) 
    {
      i2c_stop();
      i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
      i2c_write(0x00); // Co = 0, D/C = 0
      bytesOut = 1;
    }
    i2c_write(pgm_read_byte(commands++));
    bytesOut++;
  }    
  i2c_stop();
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_update
  Description  :  transfers the local copy to the display
  Argument(s)  :  None.
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_update (void)
{
  uint16_t i;
  i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
  i2c_write(0x00);
  i2c_write(SSD1306_PAGEADDR);
  i2c_write(0);
  i2c_write(0xff);
  i2c_write(SSD1306_COLUMNADDR);
  i2c_write(0);
  i2c_write(DISPLAYSIZEX - 1);
  i2c_stop();
  _delay_ms(10);
  
  i2c_start(SSD1306_DEFAULT_ADDRESS+I2C_WRITE);
  i2c_write(0x40);
  for (i=0; i<(DISPLAYSIZEX*DISPLAYSIZEY/8); i++)
  {
      #ifdef DISPLAYROTATE
      i2c_write(framebuffer[DISPLAYSIZEX*DISPLAYSIZEY/8-i-1]);
      #else
      i2c_write(framebuffer[i]);
      //i2c_write(i);
      #endif
  }
  i2c_stop();
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_clear
  Description  :  Clears the display
  Argument(s)  :  None.
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_clear ( void )
{
    DISPLAY_clearbuffer();
    DISPLAY_update();
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_init
  Description  :  LCD controller initialization.
  Argument(s)  :  contrast value VOP
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_init ( void )
{
  i2c_init();

  static const uint8_t PROGMEM init1[] = 
  { 
    SSD1306_DISPLAYOFF,         
    SSD1306_SETDISPLAYCLOCKDIV, 
    0x80, // the suggested ratio 0x80
    SSD1306_SETMULTIPLEX
  }; 
  SSD1306_sendCommandList(init1, sizeof(init1));

  SSD1306_sendCommand1(DISPLAYSIZEY - 1);

  static const uint8_t PROGMEM init2[] = 
  {
    SSD1306_SETDISPLAYOFFSET, 
    0x0,                      
    SSD1306_SETSTARTLINE | 0x0, 
    SSD1306_CHARGEPUMP
  };      
  SSD1306_sendCommandList(init2, sizeof(init2));
  
  SSD1306_sendCommand1((VCCSTATE == SSD1306_EXTERNALVCC) ? 0x10 : 0x14);

  static const uint8_t PROGMEM init3[] = 
  {
    SSD1306_MEMORYMODE, 
    0x00, // 0x0 act like ks0108
    SSD1306_SEGREMAP | 0x1,
    SSD1306_COMSCANDEC
  };
  SSD1306_sendCommandList(init3, sizeof(init3));  
  
  uint8_t comPins = 0x02;
  uint8_t contrast = 0x8F;

  if ((DISPLAYSIZEX == 128) && (DISPLAYSIZEY == 32)) 
  {
    comPins = 0x02;
    contrast = 0x8F;
  } 
  else 
  if ((DISPLAYSIZEX == 128) && (DISPLAYSIZEY == 64)) 
  {
    comPins = 0x12;
    contrast = (VCCSTATE == SSD1306_EXTERNALVCC) ? 0x9F : 0xCF;
  } 
  else 
  if ((DISPLAYSIZEX == 96) && (DISPLAYSIZEY == 16)) 
  {
    comPins = 0x2; // ada x12
    contrast = (VCCSTATE == SSD1306_EXTERNALVCC) ? 0x10 : 0xAF;
  } 
  else 
  {
    // Other screen varieties -- TBD
  }  
  SSD1306_sendCommand1(SSD1306_SETCOMPINS);
  SSD1306_sendCommand1(comPins);
  SSD1306_sendCommand1(SSD1306_SETCONTRAST);
  SSD1306_sendCommand1(contrast);
  
  SSD1306_sendCommand1((VCCSTATE == SSD1306_EXTERNALVCC) ? 0x22 : 0xF1);
    
  SSD1306_sendCommand1(SSD1306_SETPRECHARGE); // 0xd9
  SSD1306_sendCommand1((VCCSTATE == SSD1306_EXTERNALVCC) ? 0x22 : 0xF1);
  static const uint8_t PROGMEM init5[] = 
  {
    SSD1306_SETVCOMDETECT, 
    0x40,
    SSD1306_DISPLAYALLON_RESUME, 
    SSD1306_NORMALDISPLAY,       
    SSD1306_DEACTIVATE_SCROLL,
    SSD1306_DISPLAYON
  }; 
  SSD1306_sendCommandList(init5, sizeof(init5));

  // Turn display back on
  SSD1306_sendCommand1(SSD1306_DISPLAYON);
    
  DISPLAY_clear();
   
}


/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_setpixel
  Description  :  Sets the pixel at xy location
  Argument(s)  :  x 
                  y 
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_setpixel(uint16_t x, uint16_t y)
{
  if ((x < DISPLAYSIZEX) && (y < DISPLAYSIZEY))
  {
    framebuffer[(uint16_t) x+DISPLAYSIZEX*(y/8)] |= (1 << (y % 8));
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_clearpixel
  Description  :  Clears the pixel at xy location
  Argument(s)  :  x 
                  y 
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_clearpixel(uint16_t x, uint16_t y)
{
  if ((x < DISPLAYSIZEX) && (y < DISPLAYSIZEY))
  {
    framebuffer[(uint16_t) x+DISPLAYSIZEX*(y/8)] &= ~(1 << (y % 8));
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_drawpixel
  Description  :  Sets the pixel at xy location
  Argument(s)  :  x 
                  y 
                  color
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_drawpixel(uint16_t x, uint16_t y, uint8_t color)
{
  if ((x < DISPLAYSIZEX) && (y < DISPLAYSIZEY))
  {
    if (color)
    {
      DISPLAY_setpixel(x, y);
    } 
    else
    {
      DISPLAY_clearpixel(x, y);
    }
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_hLine
  Description  :  draw a horizontal line
  Argument(s)  :  x1, x2, y, color
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_hLine(uint16_t x1, uint16_t x2, uint16_t y, uint8_t color)
{
  uint16_t x;
  if (x1 > x2)
  {
    x  = x1;
    x1 = x2;
    x2 = x;
  }
  if ((x2 < DISPLAYSIZEX) && (y < DISPLAYSIZEY))
  {
    for (x=x1; x<=x2; x++)
    {
      DISPLAY_drawpixel(x, y, color);
    }
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_vLine
  Description  :  draw a vertical line
  Argument(s)  :  x, y1, y2, color
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_vLine(uint16_t x, uint16_t y1, uint16_t y2, uint8_t color)
{
  uint16_t y;
  if (y1 > y2)
  {
    y  = y1;
    y1 = y2;
    y2 = y;
  }
  if ((x < DISPLAYSIZEX) && (y2 < DISPLAYSIZEY))
  {
    for (y=y1; y<=y2; y++)
    {
      DISPLAY_drawpixel(x, y, color);
    }
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_Rect
  Description  :  draw a rectangle
  Argument(s)  :  x1, y1, w, h, color
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_Rect(uint16_t x1, uint16_t y1, uint16_t w, uint16_t h, uint8_t color)
{
  uint16_t x2, y2, y;
  x2 = x1+w;
  y2 = y1+h;

  if (x2 >= DISPLAYSIZEX)
  {
    x2 = DISPLAYSIZEX-1;
  }

  if (y2 >= DISPLAYSIZEY)
  {
    y2 = DISPLAYSIZEY-1;
  }

  if ((x1 < DISPLAYSIZEX) && (y1 < DISPLAYSIZEY))
  {
    for (y=y1; y<=y2; y++)
    {
      DISPLAY_hLine(x1, x2, y, color);
    }      
  }
}


/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_putchar
  Description  :  puts a single character onto LCD
  Argument(s)  :  x - range: 0 to 83
                  y - range: 0 to 47
                  ch - character
                  attr - attribute 0-normal, 1-inverse, 2-underline
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_putchar(uint8_t x0, uint8_t y0, char ch, uint8_t attr)
{
  uint8_t yd, ym, i, fontbyte;
  uint16_t m;
  yd = y0/8;
  ym = y0%8;
  for (i=0; i<6; i++)
  {
    fontbyte = pgm_read_byte(&smallFont[(uint8_t)ch][i]);
    switch (attr)
    {
      case  0:
          break;
      case  1: 
          fontbyte ^= 0xff;
          break;
      case  2: 
          fontbyte |= 0b10000000;
          break;
    }

    if ((x0+i)<DISPLAYSIZEX)
    {
      m = (uint16_t) x0+i+DISPLAYSIZEX*(yd);
      framebuffer[m] &= ~(0xff << ym);
      framebuffer[m] |= (fontbyte << ym);
      if ((y0<(DISPLAYSIZEY-8)) && (ym != 0))
      {
        m = (uint16_t) x0+i+DISPLAYSIZEX*(yd+1);
        framebuffer[m] &= ~(0xff >> (8-ym));
        framebuffer[m] |= (fontbyte >> (8-ym));
      }
    }
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_print
  Description  :  prints a string
  Argument(s)  :  x - range: 0 to 83
                  y - range: 0 to 47
                  *ch - pointer t string
                  attr - attribute 0-normal, 1-inverse, 2-underline
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_print(uint8_t x, uint8_t y, char *ch,uint8_t attr)
{
  while (*ch)
  {
    DISPLAY_putchar(x, y, *ch, attr);
    ch++;
    x += 6;
  }
}

void DISPLAY_print_p(uint8_t x, uint8_t y, const char *ch,uint8_t attr)
{
  char c;
  while ((c = pgm_read_byte(ch)))
  {
    DISPLAY_putchar(x, y, c, attr);
    ch++;
    x += 6;
  }
}

/*--------------------------------------------------------------------------------------------------
  Name         :  DISPLAY_scroll
  Description  :  softscrolls the framebuffer
  Argument(s)  :  dy - range: 0 to +/-47
  Return value :  None.
--------------------------------------------------------------------------------------------------*/
void DISPLAY_scroll(int8_t dy)
{
  int8_t y1;
  uint8_t  x, y, dy1, dy8, b1, b2;
  if (dy>0)
  {
    dy8 = dy/8;
    dy1 = dy%8;
    for (x=0; x<DISPLAYSIZEX; x++)
    {
      for (y=0; y<(DISPLAYSIZEY/8); y++)
      {
        y1=y+dy8;
        if (y1<(DISPLAYSIZEY/8))
        {
          b1 = framebuffer[x + DISPLAYSIZEX*y1];
        }
        else
        {
          b1=0;
        }
        if ((y1+1)<(DISPLAYSIZEY/8))
        {
          b2 = framebuffer[x + DISPLAYSIZEX*(y1+1)];
        }
        else
        {
          b2=0;
        }
        framebuffer[x + DISPLAYSIZEX*(y)] = (b1 >> dy1) | (b2 << (8-dy1));
      }
    }
  } 
  else
  {
    dy8 = abs(dy)/8;
    dy1 = abs(dy)%8;
    for (x=0; x<DISPLAYSIZEX; x++)
    {
      for (y=0; y<(DISPLAYSIZEY/8); y++)
      {
        y1=(DISPLAYSIZEY/8)-y-dy8-1;
        if (y1>=0)
        {
          b1 = framebuffer[x + DISPLAYSIZEX*y1];
        }
        else
        {
          b1=0;
        }
        if ((y1-1)>=0)
        {
          b2 = framebuffer[x + DISPLAYSIZEX*(y1-1)];
        }
        else
        {
          b2=0;
        }
        framebuffer[x + DISPLAYSIZEX*((DISPLAYSIZEY/8)-y-1)] = (b1 << dy1) | (b2 >> (8-dy1));
      }
    }
  }
}
